module Interpreter
using CUDA
using ..ExpressionProcessing

export interpret

"Interprets the given expressions with the values provided.
# Arguments
 - expressions::Vector{ExpressionProcessing.PostfixType} : The expressions to execute in postfix form
 - variables::Matrix{Float64} : The variables to use. Each column is mapped to the variables x1..xn
 - parameters::Vector{Vector{Float64}} : The parameters to use. Each Vector contains the values for the parameters p1..pn. The number of parameters can be different for every expression
"
function interpret(expressions::Vector{ExpressionProcessing.PostfixType}, variables::Matrix{Float64}, parameters::Vector{Vector{Float64}})
	# TODO: 
	#      create CUDA array and fill it with the variables and parameters
	#      create CUDA array for calculation results
	variableRows = size(variables, 1)
	cudaVars = CuArray(variables)
	
	paramRows = get_max_parameter_rows(parameters)
	p1aramCols = length(parameters)
	cudaParams = CuArray{Float64}(undef, p1aramCols, paramRows) # length(parameters) == number of expressions
	# TODO: Fill cudaParams

	# TODO: Move CuArray(expression[i]) outside the loop for a more efficient transfer to GPU but leave kernel signature as is
	for i in eachindex(expressions)
		cudaExpr = CuArray(expressions[i])
		kernel = @cuda launch=false interpret_expression(cudaExpr, cudaVars, cudaParams, i)
		config = launch_configuration(kernel.fun)
		threads = min(variableRows, config.threads)
		blocks = cld(variableRows, threads)

		kernel(cudaExpr, cudaVars, cudaParams, i; threads, blocks)
	end
end

function interpret_expression(expression, variables, parameters, exprIndex::Int)
	#TODO Implement interpreter
end

"Retrieves the number of entries for the largest inner vector"
function get_max_parameter_rows(params::Vector{Vector{T}})::Int where T
	maxLength = 0
	for i in eachindex(params)
		if length(params) > maxLength
			maxLength = length(params)
		end
	end

	return maxLength
end



@deprecate InterpretExplicit!(op::Operator, x, y) interpret_expression(expression, variables, parameters, exprIndex::Int)
# Kernel
function InterpretExplicit!(op::Operator, x, y)
	index = (blockIdx().x - 1) * blockDim().x + threadIdx().x
	stride = gridDim().x * blockDim().x

	if op == Add
		# @cuprintln("Performing Addition") # Will only be displayed when the GPU is synchronized
		for i = index:stride:length(y)
			@inbounds y[i] += x[i]
		end
		return
	elseif op == Subtract
		# @cuprintln("Performing Subtraction") # Will only be displayed when the GPU is synchronized
		for i = index:stride:length(y)
			@inbounds y[i] -= x[i]
		end
		return
	end
end

end